"""
    Code for measuring MQTT-SN elapsed time over 5G using Cloudified PLC
    Closed Control Loop Latency data is exported to a CSV file
"""

import csv
import threading
import time
from datetime import datetime
import sys, os
import logging
from pathlib import Path

sys.path.append(os.path.join(Path(__file__).parents[0], "mqttsnclient"))
# sys.path.append(os.path.join(sys.path[0], "mqttsnclient"))

from mqttsnclient.MQTTSNclient import Callback
from mqttsnclient.MQTTSNclient import Client
from mqttsnclient.MQTTSNclient import publish
import mqttsnclient.MQTTSN as MQTTSN

FORMAT = '%(asctime)s, %(levelname)s, %(filename)s:%(lineno)d, %(funcName)s(), %(message)s'
logging.basicConfig(format=FORMAT, level=logging.INFO)

myLogger = logging.getLogger()

print(sys.version_info)


# message_q = queue.Queue()

#Broker address (Edge cloud)
host = "192.168.100.34"
#Port for RSMB
port = 1884
#Topic for publishing
topic_pub = "ec"
#Topic for subscribing
topic_sub = "ur"
#Client
client_name = "edge"
#Text for the 1300 bytes payload
text1500 = ""

#We ask the user fot these imput variables
n_packets = int(input("Number of packets: "))
qos = int(input("QoS {0,1]: "))
bw_load = int(input("Network load %: "))
packet_size = input("Payload size: \na)2 bytes\nb)1300 bytes\n: ")
packet_size = packet_size.lower()
#If the packet size is 1500 bytes, we generate that payload
if packet_size == "b":
    n_bytes = 1300
    with open('1500bytes.txt', 'r') as file:
        # Read all the lines of the file into a list
        lines = file.readlines()
        # Join the lines into a single string
        text1500 = ''.join(lines)
else:
    n_bytes = 2



class MyCallback(Callback):

    def on_message(self, client, TopicId, Topicname, payload, qos, retained,
                   msgid):
        m= "Arrived" +" topic  " +str(TopicId)+ "message " +\
           str(payload) +"  qos= " +str(qos) +" ret= " +str(retained)\
           +"  msgid= " + str(msgid)
        
        #If the message is 2 bytes payload
        if n_bytes == 2:
            #We take the payload of the packet to int
            idx_packet = int(payload)
        
        #If the message is 1300 bytes payload
        elif n_bytes == 1300:
            #We split the message
            rx = payload.split(";")
            #After we split the message, we get some null bytes at the end so we have to remove them
            # so we can later convert the str index of the packet to int
            # If we dont do it, it crashes
            idx_packet = rx[1].strip("\x00\x00\x00\x00\x00\x00\x00\x00")
            idx_packet = int(idx_packet)
        #We stamp the time
        rx_time[idx_packet] = datetime.now()

        print("------------------------------")
        print("MSG RX: ", payload)
        print("Topic: ", Topicname.decode('utf8'))
        print("------------------------------\n")

        #myLogger.info(m)
        #myLogger.info("got the message {}".format(payload))
        # message_q.put(payload)
        return True
    


#We create the dicts to store the time
tx_time = {}
for i in range (0,n_packets):
    tx_time[i] = datetime.now()

rx_time = {}
for i in range (0,n_packets):
    rx_time[i] = datetime.now()

dif_time = {}
for i in range (0,n_packets):
    dif_time[i] = datetime.now()


#Filename to store the data
filename = "5G_UR5_CCLL_MQTT-SN_QoS"+str(qos)+"_"+str(n_bytes)+"bytes_load"+str(bw_load)+".csv"

#Function for subscribing
def subscription():
    try:
        client.subscribe(topic_sub, qos=qos)
        client.loop_start()
        while True:
            time.sleep(10)

    except BaseException as e:
        client.loop_stop()
        client.disconnect()
        raise e


#We start the client
client = Client(client_name)
client.message_arrived_flag = False
client.registerCallback(MyCallback())
myLogger.info("threads {}".format(threading.active_count()))
myLogger.info("connecting {}".format(host))
client.connected_flag = False
#We connect to the RSMB
client.connect(host, port)

#We look for CONNACK msg
client.lookfor(MQTTSN.CONNACK)

try:
    if client.waitfor(MQTTSN.CONNACK) == None:
        myLogger.info("connection failed")
        raise SystemExit("no Connection quitting")
except Exception as e:
    logging.exception(e)
    myLogger.info("connection failed")
    raise SystemExit("no Connection quitting")

#We publish the messages
try:
    thread_sub = threading.Thread(target=subscription)
    thread_sub.start()
    time.sleep(2)

    for i in range(0, n_packets):
        
        #We study what size is the payload in order to send the data
        if n_bytes == 2:
            msg = str(i)
            payload_pub = msg
        
        if n_bytes == 1300:
            payload_pub = text1500+";"+str(i)
            msg = "...;"+str(i)

        
        tx_time[i] = datetime.now()
        client.publish(topic_pub, payload_pub, qos=qos)
        print("------------------------------")
        print("MSG PUBLISHED: ", msg)
        print("Topic: ", topic_pub)
        print("------------------------------\n")
        time.sleep(0.075)
        payload_pub = ""

    #We sleep for a while before computing the time difference
    print("Computing latency...")
    time.sleep(7)


    with open(filename, "w+", newline='') as file:
        writer = csv.writer(file)
        #We write the header
        writer.writerow(["TX Time", "RX Time", "Diff Time"])
        for i in range(0,n_packets):
            time_dif = rx_time[i] - tx_time[i]
            seconds = time_dif.total_seconds()
            milliseconds = seconds*1000
            print("----------------------------")
            print("PACKET ", i, " LATENCY IS: ", milliseconds, " millis")
            print("----------------------------")
            #Line to add to csv
            row_data = [rx_time[i], tx_time[i], milliseconds]
            #We store it in the CSV
            writer.writerow(row_data)

except BaseException as e:
    client.loop_stop()
    client.disconnect()
    raise e
