"""
    Code for measuring MQTT elapsed time over 5G using Cloudified PLC
    Closed Control Loop Latency data is exported to a CSV file
"""

#Libraries
from paho.mqtt import client as mqtt_client
import time
import datetime
from datetime import datetime
from threading import Thread
import csv

#Broker Address (NOKIA 7 EdgeCloud)
broker_address = "192.168.100.34"
#Broker port
broker_port = 1883
#Publishing topic
topic_pub = "request"
#Subscribing topic
topic_sub = "respond"
#QoS level
qos_level = -1 
#Var for storing the load of the network
load = 0
#Var for storing the size of the packet
packet_size = 0
n_bytes = 0
#Var for saving the 1500 bytes packet
text1500 = ""
#Number of packets
n_packets = 0
#Filename
filename = ""

#Dict to store the times when the data is received
rx_time = {}
for i in range(0,n_packets):
    rx_time[i] = datetime.now()

#Dict to store the times when the data is sent
sent_time = {}
for i in range(0,n_packets):
    sent_time[i] = datetime.now()

#Dictionary for time difference
diff_time = {}
for i in range(0,n_packets):
    diff_time[i] = datetime.now()


#MQTT credentials
username = "testbed"
passwd = "1234"

#Function for connecting the broker
def connect_mqtt() -> mqtt_client:
    def on_connect(client, userdata, flags, rc):
        if rc == 0:
            print("Connected to MQTT Broker!\n")
        else:
            print("Failed to connect, return code %d\n", rc)

    client = mqtt_client.Client()
    client.username_pw_set(username, passwd)
    client.on_connect = on_connect
    client.connect(broker_address, broker_port)
    return client


#PUBLISH function
def publish(client: mqtt_client.Client):
    global sent_time, qos_level

    #We sent from 0 to 99
    for i in range(0,n_packets):
        #If we chose to send 1500 bytes...
        if packet_size == "b":
            tx_msg = text1500+";"+str(i)
        #Id we chose to send 2 bytes
        else:
            tx_msg = str(i)
        #We add the time when the message is sent to the dict
        sent_time[i] = datetime.now()
        #We publish the message
        result = client.publish(topic_pub,tx_msg, qos=qos_level)
        #State code
        status = result[0]
        if status == 0:
            print("Published message: " + str(i))
            print("Topic: " + topic_pub)
        else:
            print(f"Failed to send message to topic {topic_pub}")

        time.sleep(0.015)


#SUBSCRIBER function
def subscribe(client: mqtt_client.Client):
    global qos_level
    def on_message(client, userdata, msg):
        global rx_time
        #Message received
        rx_msg = msg.payload.decode()
        #If we choose the 1500 bytes, we have to split the message to know the packet index
        if packet_size == "b":
            rx_msg = rx_msg.split(";")
            #The index is the second item in the list
            idx_msg = int(rx_msg[1])

        else:
            idx_msg = int(rx_msg)
        #We get the timestamp when receiving the message
        rx_time[idx_msg] = datetime.now()
        print("RECEIVED PACKET: ", idx_msg)
        #print("SENT PACKET ", idx_msg, " AT ", rx_time[idx_msg])
        #print("RECEIVED PACKET ", idx_msg, " AT ", rx_time[idx_msg])
        

    client.subscribe(topic_sub, qos=qos_level)
    print("Subscribed to topic: " + topic_sub)
    client.on_message = on_message

#We create this function so we can use a thread for subscribing (otherwise the loop_forever() dont let us continue)
def subscriber(client):
    subscribe(client)
    client.loop_forever()

#Main function
def run():

    global qos_level, load, packet_size, text1500, n_bytes, n_packets, filename
    #We specify the QoS
    qos_level = int(input("QoS level: "))
    print("QoS ", qos_level, " CHOSEN")
 
    #We specify the number of packets
    n_packets = int(input("NUMBER OF PACKETS: "))
    print(n_packets, " PACKETS WILL BE SENT")
    
    #We ask for the load of the network
    load = int(input("Load of the network in %: "))
    #We ask for the packet size
    packet_size = input("Size of the payload: \nA) 1-2 Bytes\nB) 1300 bytes\n:")
    packet_size = packet_size.lower()
    #If the packet size is 1500 bytes, we generate that payload
    if packet_size == "b":
        n_bytes = 1300
        with open('1500bytes.txt', 'r') as file:
            # Read all the lines of the file into a list
            lines = file.readlines()
            # Join the lines into a single string
            text1500 = ''.join(lines)
    else:
        n_bytes = 2


    #Filename specifies the QoS, payload size and load on the network of the scenario being used
    filename = "5G_UR5_CCLL_MQTT_QoS_"+str(qos_level)+"_"+str(n_bytes)+"bytes_load"+str(load)+".csv"
    
    #We connect
    client = connect_mqtt()
    #Whe create a thread for subscribing, arguments -> client
    thread1 = Thread(target=subscriber, args = (client,))
    thread1.start()
    #Whe create a thread for the study of packet loss
    #thread2 = Thread(target=handle_lost)
    #thread2.start()
    #If we dont put this sleep, the first packet can be lost due to the sync of the thread and publish
    time.sleep(1)
    #We open our CSV file
    with open(filename, "w+", newline='') as file:
        writer = csv.writer(file)
        #We write the header
        writer.writerow(["TX Time", "RX Time", "Diff Time"])
        #We study 20 iterations
    
        publish(client)
        #We sleep 5 seconds because we should wait for all the packets to be finished
        time.sleep(7)
        #Once we have sent and received all the packets of the iteration we compute and store the times
        for i in range(0,n_packets):
            #We compute the closed loop delay in milliseconds
            diff_time[i] = rx_time[i] - sent_time[i]
            seconds = diff_time[i].total_seconds()
            milliseconds = seconds*1000
            #We store the data
            row_data = [rx_time[i], sent_time[i], milliseconds]
            #We store it in the CSV
            writer.writerow(row_data)
            #print(i,"Sent time: ", sent_time[i], "RX time: ", rx_time, "Diff time: ", diff_time[i])
             #We get results on packet loss
    
    client.disconnect()

if __name__ == '__main__':
    run()
