"""
    Code for measuring OPC-UA ClientServer elapsed time over 5G using Cloudified PLC
    Closed Control Loop Latency data is exported to a CSV file
"""

import asyncio
import logging
from asyncua import Client,ua
from datetime import datetime
import csv

url = "opc.tcp://10.94.132.9:4840/freeopcua/server/"
# namespace = "http://examples.freeopcua.github.io"
namespace = "OPC-UA_Server"

#Var for storing the number of packets
n_packets = 0

#Var for storing the number of lost packets
count_lost = 0
#Var to store the size of the payload
payload_size = ""

#Dict for before packet times
before_times = {}
for i in range(0,n_packets):
    before_times[i] = datetime.now()

#Dict for after packet times
after_times = {}
for i in range(0,n_packets):
    after_times[i] = datetime.now()

#Dict for elapsed  times
diff_times = {}
for i in range(0,n_packets):
    diff_times[i] = -1.0

filename = ""

async def compute_time():
    global filename, count_lost, n_packets, payload_size

    with open(filename, "w+", newline='') as file:
        writer = csv.writer(file)
        #We write the header
        writer.writerow(["TX Time", "RX Time", "Diff Time"])

        for i in range(0,n_packets):

            diff_time = after_times[i]-before_times[i]
            seconds = diff_time.total_seconds()
            milliseconds = seconds*1000
            diff_times[i] = milliseconds
            #We check if we had timeout
            #New row of the csv
            row_data = [before_times[i], after_times[i], diff_times[i]]
            #We store it in the CSV
            writer.writerow(row_data)
            print("Diff time is: ", milliseconds)




async def main():
    global filename, count_lost, n_packets, payload_size
    #We specify the BW used
    bw_used = int(input("Load of the network: "))
    #We specify the payload size
    payload_size = input("Payload size: \na) 2 bytes\nb) aprox 1300 bytes\n: ")

    payload_size = payload_size.lower()
    #If the packet size is 1500 bytes, we generate that payload
    if payload_size == "b":
        n_bytes = 1300
        with open('1500bytes.txt', 'r') as file:
            # Read all the lines of the file into a list
            lines = file.readlines()
            # Join the lines into a single string
            text1500 = ''.join(lines)
    else:
        n_bytes = 2

    #We specify the number of packets
    n_packets = int(input("Number of packets: "))
    #We specify the filename 
    filename = "5G_UR_cloop_opcua_client_server_5G_"+str(n_bytes)+"bytes_loaded_"+str(bw_used)+".csv"
    
    _logger = logging.getLogger(__name__)
    _logger.info(f"Connecting to {url} ...")
    async with Client(url=url) as client:
        # Find the namespace index (the nodes i and j indexes)
        namespaceidx = await client.get_namespace_index(namespace)
        _logger.info(f"Namespace Index for '{namespace}': {namespaceidx}")

       
        #50 times
        for i in range(0,n_packets):
            # Get the variable node for read / write
            # we have to write the same node (PLC) and variable name of this node (Temp) as in the server
            #??WE THINK THAT THE get_child() METHOD BUILDS THE REQUEST
            temperature = await client.nodes.root.get_child(
                ["0:Objects", f"{namespaceidx}:UR_Arm", f"{namespaceidx}:Temp"])
            
            #We get the data type of the variable from the server
            data_type = await temperature.read_data_type_as_variant_type()
            #We study the data type so we can be aware at the time of writing the message
            if str(data_type) == "VariantType.String":
                var_type = ua.VariantType.String
                #If the packet size is 2 bytes we send the number of the packet
                if n_bytes == 2:
                    payload = str(i)
                #Id the packet size is around 1300 bytes we send the full message + ; + number of packet
                elif n_bytes == 1300:
                    payload = text1500+";"+str(i)

            elif str(data_type) == "VariantType.Int64":
                payload = i
                var_type = ua.VariantType.Int64
                
                
            before_read_time = datetime.now()
            before_times[i] = before_read_time
            #We read the value of the temperature
            read_temp = await temperature.read_value()
            #We write the new value after we know the prior one
            temp_value = await temperature.write_value(payload, var_type)
            #After reading, we stamp the time
            after_write_time = datetime.now()
            
            #If the size of the payload is aprox 1500, we split the huge msg and get the number of the packet
            if n_bytes == 1300: 
                aux = read_temp.split(";")
                read_value = aux[1]
            else:
                read_value = read_temp

            print("READ VALUE IS: ", read_value)
            print("NEW VALUE FOR TEMP: ", i)

            after_times[i] = after_write_time
            #_logger.info(f"Value of Temperature ({temperature}): {temp_value}")
   
            await asyncio.sleep(0.015)             
        
        await compute_time()

        await client.disconnect()



if __name__ == "__main__":
    #logging.basicConfig(format="%(asctime)-15s %(message)s",level=logging.INFO)
    asyncio.run(main())
    
