


"""
Code used for moving the robotic arm from the Edge Cloud (Cloudified PLC)
Output: a CSV is created storing the total cycle elapsed rime of the whole routine
"""


#Libraries
from paho.mqtt import client as mqtt_client
from threading import Thread
import time
import csv

#Broker Address
broker_address = "192.168.100.34"
#Broker port
broker_port = 1883
#Message to publish
msg_pub = ""
#Publishing topic
topic_pub = "joints_val"
#Subscribing topic
topic_sub = "current_val"
#QoS level
qos_level = 0
#Current position (subscribed)
current_pos = None
#Numbr of iterations
n_times = 0
#List for storing the values of time elapsed
time_list = []

#Function for connecting the broker
def connect_mqtt() -> mqtt_client:
    def on_connect(client, userdata, flags, rc):
        if rc == 0:
            print("Robot operator has connected to MQTT Broker!\n")
        else:
            print("Failed to connect, return code %d\n", rc)

    client = mqtt_client.Client()
    #Uncomment when using authenticated communications
    client.username_pw_set(username="testbed", password="1234")
    client.on_connect = on_connect
    client.connect(broker_address, broker_port)
    return client


#PUBLISH function
def publish(client: mqtt_client.Client, topic_pub, msg):
    global qos_level
    #We publish the message
    result = client.publish(topic_pub,msg, qos=qos_level)
    status = result[0]
    if status == 0:
        """
        print("---------------------------------")
        print("Published message: " + msg)
        print("Topic: " + topic_pub)
        print("---------------------------------")
        """
    else:
        print(f"Failed to send message to topic {topic_pub}")
    #10 ms of rate
    #IT CAN BE CHANGED
    time.sleep(0.01)



#Function used to convert a string list to normal list (form "[12, 45, 23]" to [12,45,23])
def convert_str_list(data_input: str):
    data_output = data_input.replace(" ", "")
    data_output = data_output.replace("[", "")
    data_output = data_output.replace("]", "")
    output = data_output.split(",")

    converted_list = []
    for i in range(0, len(output)):
        converted_list.append(float(output[i]))
    return converted_list

#Function used for waiting to the final position
def wait_pos(dest_pos):

    #Our threshold is 0.01*pi/180 = 0.000175 (this is our error of position in radians of every joint)
    #This is because the robot joint usually changes the 0.01 degree, it is not stable 
    #We will do it with the threshold 5 times
    threshold = 0.000175*5
    
    while True:
        time.sleep(0.001)
        #We get the current OPCUA position of the joints and convert it to list
        #print("CURRENT POS B4 CONVERT: ", current_pos)
        current_mqtt_position = convert_str_list(current_pos)
        #print("CURRENT POS: ", current_mqtt_position)
        #print("DESTINATION POS: ", dest_pos)
        #Flag to exit to the while True level
        flag = True

        for i in range(0,len(dest_pos)):
            result = abs(dest_pos[i]-current_mqtt_position[i])
            if result > threshold:
                flag = False
                break
        
        if flag == True:
            #print("reach")
            break
        #print("not reach")



#Function used for executing the movement
#The list has to be be edited in order to change the movement routine
def move_1(client):
        global time_list
        #Array of all positions
        all_positions = []
        all_positions.append([0.125664, -2.07083, -2.20086, -0.39235, 1.459793, 0.890642])
        all_positions.append([0.125664, -1.80903, -2.20086, -0.39235, 1.459793, 0.890642])
        all_positions.append([-1.51058, -1.80903, -2.20086, -0.39235, 1.459793, 0.890642])
        all_positions.append([-1.51058, -1.80903, -1.102, -0.39235, 1.459793, 0.890642])
        all_positions.append([-1.51058, -1.80903, -1.102, -0.23754, 1.459793, 0.890642])

        start_time = time.time()

        for i in range(0,len(all_positions)):
            publish(client, topic_pub,"[{}, {}, {}, {}, {}, {}]".format(all_positions[i][0],all_positions[i][1],all_positions[i][2],all_positions[i][3],all_positions[i][4],all_positions[i][5]))
            #print("movimiento ", i , " ASCENDENTE")
            wait_pos(all_positions[i])
            #time.sleep(1)

        for i in range(0,len(all_positions)):
            publish(client, topic_pub,"[{}, {}, {}, {}, {}, {}]".format(all_positions[len(all_positions)-i-1][0],all_positions[len(all_positions)-i-1][1],all_positions[len(all_positions)-i-1][2],all_positions[len(all_positions)-i-1][3],all_positions[len(all_positions)-i-1][4],all_positions[len(all_positions)-i-1][5]))
            #print("movimiento ", i , " DESCENDENTE")
            wait_pos(all_positions[len(all_positions)-i-1])
            #time.sleep(1)
        end_time = time.time()
        elapsed_time = end_time - start_time
        print("ELAPSED TIME MQTT: ", round(elapsed_time,3), " SECONDS")
        #We add the elapsed time
        time_list.append(elapsed_time)



#SUBSCRIBER function
def subscribe(client: mqtt_client.Client):
    global qos_level
    def on_message(client, userdata, msg):
        global current_pos
        #Message received
        rx_msg = msg.payload.decode()
        current_pos = rx_msg
        rx_topic = msg.topic
        """
        print("---------------------------------")
        print("Received message " + rx_msg)
        print("Topic: " + rx_topic)
        print("---------------------------------")
        """
    client.subscribe(topic_sub, qos=qos_level)
    print("Subscribed to topic: " + topic_sub)
    client.on_message = on_message

#We create this function so we can use a thread for subscribing (otherwise the loop_forever() dont let us continue)
def subscriber(client):
    subscribe(client)
    client.loop_forever()


#Main function
def run():
    global msg_pub, qos_level, n_times
    #We connect
    client = connect_mqtt()
    #Whe create a thread for subscribing, arguments -> client
    thread1 = Thread(target=subscriber, args = (client,))
    thread1.start()
    #If we dont put this sleep, the first packet can be lost due to the sync of the thread and publish
    time.sleep(1)
    #We ask for the times
    #n_times = int(input("Number of times: "))
    n_times = 5
    #Filename
    filename = "TIME_ELAPSED_MOVEROUTINE_MQTT_"+str(n_times)+"_TIMES.csv"
    """"
    while True:
       #We move the arm
       move_1(client)
    """
    for i in range(0,n_times):
        print("========================")
        print("Iteration ",i)
        move_1(client)
        print("========================")


    with open(filename, 'w', newline='') as csvfile:
        writer = csv.writer(csvfile)
        # Write the headers
        writer.writerow(['Iteration', 'ElapsedTime'])
        # Write the list elements
        for index, elapsed_time in enumerate(time_list, start=1):
            writer.writerow([index, elapsed_time])
        
    print("CSV file created and written!")
    print("Test has ended!")
 
if __name__ == '__main__':
    run()
