"""
    Code for measuring MQTT elapsed time over 5G using Cloudified PLC
    Robotic cell side (CCLL data is not gathered here, it is collected in Cloud's side)
"""

#Libraries
from paho.mqtt import client as mqtt_client


#Broker address (EDGE CLOUD NOKIA 7)
broker_address = "192.168.100.34"
#Broker port
broker_port = 1883
#Publishing topic
topic_pub = "response"
#Subscribing topic
topic_sub = "request"
#QoS level
qos_level = -1


#MQTT credentials
username = "testbed"
passwd = "1234"

#Function for connecting the broker
def connect_mqtt() -> mqtt_client:
    def on_connect(client, userdata, flags, rc):
        if rc == 0:
            print("Connected to MQTT Broker (NOKIA 7 Edge Cloud)!\n")
        else:
            print("Failed to connect, return code %d\n", rc)
    
    client = mqtt_client.Client()
    client.username_pw_set(username, passwd)
    client.on_connect = on_connect
    client.connect(broker_address, broker_port)
    return client


#PUBLISH function (we specify the type of client as a MQTT client)
def publish(client: mqtt_client.Client,idx_rx):
    global qos_level
    tx_msg = idx_rx
    result = client.publish(topic_pub,tx_msg,qos=qos_level)
    #Status
    status = result[0]
    if status == 0:
        print("Published message: " + tx_msg)
        print("Topic: " + topic_pub)
    else:
        print(f"Failed to send message to topic {topic_pub}")

    print("\n") 


#SUBSCRIBE function
def subscribe(client: mqtt_client.Client):
    global qos_level
    def on_message(client, userdata, msg):
        #What we receive
        rx_msg = msg.payload.decode()
        print("RECEIVED: ",rx_msg)
        #We publish the message we RX (we answer with it to node 1)
        publish(client,rx_msg)
        
    client.subscribe(topic_sub, qos=qos_level)
    print("Subscribed to topic: " + topic_sub)
    client.on_message = on_message


def run():

    global qos_level
    #We define the qos level
    qos_level = int(input("QoS level: "))

    client = connect_mqtt()
    subscribe(client)
    client.loop_forever()


if __name__ == '__main__':
    run()
