"""
    Code for measuring OPC-UA elapsed time over 5G using Cloudified PLC
    Robotic cell side (CCLL data is not gathered here, it is collected in Cloud's side)
"""

import asyncio
import logging
import random
import sys

from asyncua import Server,ua

async def main():
    _logger = logging.getLogger(__name__)
    # set up our server
    server = Server()
    await server.init()
    server.set_endpoint("opc.tcp://0.0.0.0:4840/freeopcua/server/")

    #We ask for the data type
    data_type = int(input("Data type for the variable Test: \n1)Boolean\n2)Int64\n3)UInt64\n4)Float\n5)String\n: "))
    
    #We change the data type
    #Boolean (ID 1)
    if data_type == 1:
        var_type = ua.VariantType.Boolean
        default = False
    
    #Int64 (ID 8)
    elif data_type == 2:
        var_type = ua.VariantType.Int64
        default = 0

    #UInt64 (ID 9)
    elif data_type == 3:
        var_type = ua.VariantType.UInt64
        default = 0
    
    #Float (ID 10)
    elif data_type == 4:
        var_type = ua.VariantType.Float
        default = 1.0
    
    #String (ID 12)
    elif data_type == 5:
        var_type = ua.VariantType.String
        default = ""
    
    #Namespace
    uri = "OPC-UA_Server"
    idx = await server.register_namespace(uri)
    #Log messages
    _logger.info("idx is {}".format(idx))

    # server.nodes, contains, links to very common nodes like objects and root
    #We creatre a Node inside the namespace of this server
    plc_node = await server.nodes.objects.add_object(idx, "UR_Arm")
    #Log message
    _logger.info("myobj is {}".format(plc_node))
    #We add one variable to the node
    temperature = await plc_node.add_variable(idx, "Temp", default, var_type)
    #Log message
    _logger.info("myvar is {}".format(temperature))
    # Set MyVariable to be writable by clients
    await temperature.set_writable()


    _logger.info("Starting server!")
    async with server:
        try:
            while True:
                sleep_seconds = random.randint(1, 3)
                await asyncio.sleep(sleep_seconds)
                #We read the variable value
                current_value = await temperature.read_value()
                print("Current value: ", current_value)
                #We use .info() beacuse a logger doesnt support a print
                #_logger.info("Set value of %s to %d", temperature, int(current_value))
                #We write the variable value
                #await temperature.write_value(current_value)

        #We finish the server if there is an exception (keyboard)        
        except (KeyboardInterrupt, SystemExit):
            server.stop()
            sys.exit()

if __name__ == "__main__":
    #If we set level = logging.DEBUG there is a lot of information
    #If we set level = logging.INFO we get less info
    logging.basicConfig(format="%(asctime)-15s %(message)s",
                        level=logging.INFO)
    #Main run() of the server
    asyncio.run(main(), debug=True)
