
"""
    Code for measuring OPC-UA elapsed time over 5G using Cloudified PLC
    Robotic cell side (CCLL data is not gathered here, it is collected in Cloud's side)
"""

import asyncio
import sys

from asyncua import Server,ua,Node

#Period of PubSub (it has to be the same as the node1)
period_ms = 0
#Initial value of the variable 1
value_var1 = ""

class SubHandler:

    #Function initialization (constructor)
    def __init__(self, aux_value: Node,  variable_pub: Node):
        self.aux_value = aux_value
        self.variable_pub = variable_pub

    #Function that tells us about the change in data
    async def datachange_notification(self, node, val, attr):
        global value_var1
        print("NODE 1 HAS MADE A CHANGE IN REQUEST")
        #We see what is the new data value (it changed)
        #print("New data change event:")
        #print("Node: ", node)
        #print("Value: ", val)
        #We get the current value of the variable 1 and store it 
        value_var1 = val
        #We write the variable 2 with that value
        await self.variable_pub.write_value(value_var1, ua.VariantType.String)
        print("NODE 2 --> CHANGES VALUE OF REQUEST TO RESPONSE")



#Function for the SUBSCRIPTION
async def subscription(server: Server, variable_sub:Node, variable_pub:Node):
    print("======================= WE ARE INSIDE SUBSCRIPTION() =================")
    #We create a SUBSCRIPTION.
    sub = await server.create_subscription(period_ms, SubHandler(variable_sub, variable_pub))
    #We subscribe to the value change of the variable Var1
    handle_var = await sub.subscribe_data_change(variable_sub)
    print("NODE 2 ---> SUBSCRIBED TO REQUEST")
    return sub, handle_var


async def main():
    global period_ms
    #_logger = logging.getLogger(__name__)
    # set up our server
    server = Server()
    await server.init()
    server.set_endpoint("opc.tcp://0.0.0.0:4840/freeopcua/server/")

    #We ask for the refreshing period
    period_ms = int(input("PubSub period {1,10,50}: "))
    #We ask for the data type
    data_type = int(input("Data type for the variable Test: \n1)Boolean\n2)Int64\n3)UInt64\n4)Float\n5)String\n: "))
    
    #We change the data type
    #Boolean (ID 1)
    if data_type == 1:
        var_type = ua.VariantType.Boolean
        default = False
    
    #Int64 (ID 8)
    elif data_type == 2:
        var_type = ua.VariantType.Int64
        default = 0

    #UInt64 (ID 9)
    elif data_type == 3:
        var_type = ua.VariantType.UInt64
        default = 0
    
    #Float (ID 10)
    elif data_type == 4:
        var_type = ua.VariantType.Float
        default = 1.0
    
    #String (ID 12)
    elif data_type == 5:
        var_type = ua.VariantType.String
        default = ""
    
    #Namespace
    uri = "OPC-UA_Server"
    idx = await server.register_namespace(uri)
    #Log messages
    #_logger.info("idx is {}".format(idx))

    # server.nodes, contains, links to very common nodes like objects and root
    #We creatre a Node inside the namespace of this server
    plc_node = await server.nodes.objects.add_object(idx, "PLC")
    #Log message
    #_logger.info("myobj is {}".format(plc_node))

    #We add the following variables to the OPC-UA server:
    # Var1 -> this is the variable that Node 1 changes and Node2 detects that change (so we call it var_sub = variable to subscribe)
    # Var2 -> this is the variable that Node 2 changes after being notified of a change in Var1 (so we call it var_pub = variable to publish)
    var_sub = await plc_node.add_variable(idx, "Request", default, var_type)
    var_pub= await plc_node.add_variable(idx, "Response", default, var_type)

    # Set MyVariable to be writable by clients
    await var_pub.set_writable()
    await var_sub.set_writable()

    #We subscribe Node2 to variable var_sub (var1)
    sub_var, sub_handle_var = await subscription(server,var_sub,var_pub)

    #_logger.info("Starting server!")
    async with server:
        try:
            while True:
                await asyncio.sleep(0.5)

        except (KeyboardInterrupt, SystemExit):
            await sub_var.unsubscribe(sub_handle_var)
            await sub_var.delete()
            await server.stop()
            sys.exit()


if __name__ == "__main__":
    #If we set level = logging.DEBUG there is a lot of information
    #If we set level = logging.INFO we get less info
    #logging.basicConfig(format="%(asctime)-15s %(message)s",level=logging.INFO)
    #Main run() of the server
    asyncio.run(main(), debug=True)
