
"""
Code used for moving the robotic arm from the Edge Cloud (Cloudified PLC)
Robotic cell side
"""

import sys
sys.path.append("..")
import logging
import json

import paho.mqtt.client as mqtt 

import logging
import json

import paho.mqtt.client as mqtt 

#UR libraries API
import rtde.rtde as rtde
import rtde.rtde_config as rtde_config

from collections import namedtuple
from threading import Thread

from collections import namedtuple
from threading import Thread
import time


#We start the motion variable to none 
# before getting data from RTDE we need to have something even though it is empty
#Initial position
motion = [0,0,0,0,0,0]


# logging.basicConfig(level=logging.INFO)
ROBOT_HOST = "127.0.0.1"
ROBOT_PORT = 30004
config_filename = "control_loop_configuration.xml"


logging.getLogger().setLevel(logging.INFO)

conf = rtde_config.ConfigFile(config_filename)
state_names, state_types = conf.get_recipe("state")
actualq_names, actualq_types = conf.get_recipe("actualq")
#watchdog_names, watchdog_types = conf.get_recipe("watchdog")

con = rtde.RTDE(ROBOT_HOST, ROBOT_PORT)
con.connect()
print("Connected to RTDE")

# get controller version
con.get_controller_version()

# setup recipes
con.send_output_setup(state_names, state_types)
actual_q = con.send_input_setup(actualq_names, actualq_types)


#We apply the initial position to the joints
actual_q.input_double_register_24 = motion[0]
actual_q.input_double_register_25 = motion[1]
actual_q.input_double_register_26 = motion[2]
actual_q.input_double_register_27 = motion[3]
actual_q.input_double_register_28 = motion[4]
actual_q.input_double_register_29 = motion[5] 
# start data synchronization
if not con.send_start():
    sys.exit()

con.send(actual_q)
#Topics
topic_sub = "joints_val"
topic_pub = "current_val"

#QoS level
qos_level = 0

#Current position of the arm
current_RTDE = None

keep_running = False

#Frecuency rate for RTDE
frec_RTDE = 0


#Function used to read the data of the joints using RTDE
def read_RTDE():
    global current_RTDE
    conf = rtde_config.ConfigFile("record_configuration.xml")
    output_names, output_types = conf.get_recipe("out")

    # get controller version
    con.get_controller_version()

    # setup recipes
    if not con.send_output_setup(output_names, output_types, frequency=50):
        logging.error("Unable to configure output")
        sys.exit()

    # start data synchronization
    if not con.send_start():
        logging.error("Unable to start synchronization")
        sys.exit()
   

    keep_running = True
    while keep_running:
        try:
            
            #Getting data from the joints from RTDE
            state = con.receive_buffered()
            if state is not None:
                #We convert the data into json object
                data = json.dumps(state.__dict__) #serialize data object
                #Converting JSON data into Python readable object
                current_RTDE = json.loads(data, object_hook = lambda d : namedtuple('a', d.keys()) (*d.values()))
                #print("ESTAMOS DENTRO DEL TRY DE READ RTDE")

        except KeyboardInterrupt:
            keep_running = False
        except rtde.RTDEException:
            con.disconnect()
            sys.exit()

    con.send_pause()
    con.disconnect()


#Function used to write the joint with the values we get from the edge cloud
def write_RTDE(motion):

    actual_q.input_double_register_24 = motion[0]
    actual_q.input_double_register_25 = motion[1]
    actual_q.input_double_register_26 = motion[2]
    actual_q.input_double_register_27 = motion[3]
    actual_q.input_double_register_28 = motion[4]
    actual_q.input_double_register_29 = motion[5] 
        
    con.send(actual_q)
    print("SE HA APLICADO EL COMANDO A RTDE CORRECTAMENTE")
    print(motion)
    

#PUBLISH function
def publish(client: mqtt.Client, topic_pub, msg):
    global qos_level
    #We publish the message
    result = client.publish(topic_pub,msg, qos=qos_level)
    status = result[0]
    if status == 0:
        print("---------------------------------")
        print("Published message: " + msg)
        print("Topic: " + topic_pub)
        print("---------------------------------")

    else:
        print(f"Failed to send message to topic {topic_pub}")



def on_connect(client, userdata, flags, return_code):
    if return_code == 0:
        print("UR Robot connected to broker!")
        #We subscribe
        client.subscribe(topic_sub)
        print("ME HE SUSCRITO")
    else:
        print("could not connect to mqtt broker, return code:", return_code)


#Function for converting from list in string to normal list
def convert_str_list(data_input: str):
    data_output = data_input.replace(" ", "")
    data_output = data_output.replace("[", "")
    data_output = data_output.replace("]", "")
    output = data_output.split(",")
    converted_list = []
    for i in range(0, len(output)):
        converted_list.append(float(output[i]))
    return converted_list


def on_message(client, userdata, message):
    global motion, keep_running
    #We receive the message of the motion of the other robot

    #motion = json.loads(message.payload.decode("utf-8"), object_hook = lambda d : namedtuple('a', d.keys()) (*d.values()))
    rx_msg = message.payload.decode("utf-8")
    #
    motion_str = rx_msg
    #We convert the string list to a normal list
    try:
        print("MENSAJE RECIBIDO: ", motion_str)
        motion = convert_str_list(motion_str)
        print("SE HA CONVERTIDO A STR")
        write_RTDE(motion)
    except Exception as e:
        print("EXCEPTION: ", e)
    keep_running = True

    

def main():
    global keep_running
    time.sleep(5)
    try:
        #We create an old list to know the old position
        old_current_pos_list = ""
        while True:

            time.sleep(0.001)
            #print("ESTAMOS DENTRO DEL WHILE TRUE")
            #We store the current position
            current_pos_list = "["+str(current_RTDE.actual_q[0])+","+str(current_RTDE.actual_q[1])+"," + str(current_RTDE.actual_q[2])+"," + str(current_RTDE.actual_q[3]) + "," + str(current_RTDE.actual_q[4]) + "," + str(current_RTDE.actual_q[5])+"]"
            #If our last position (old) and the current one are different, we publish our last position
            if old_current_pos_list != current_pos_list:
                current_values = client.publish(topic_pub,current_pos_list, qos=qos_level)
                #print("SE HA PUBLICADO EL VALOR AL EDGE")
                old_current_pos_list = current_pos_list

    #We finish the server if there is an exception (keyboard)        
    except (KeyboardInterrupt, SystemExit):
        sys.exit()



client = mqtt.Client("UR5arm")
client.username_pw_set(username="testbed", password="1234")
client.on_connect=on_connect
client.on_message=on_message
client.connect("192.168.100.34", 1883)

thread_read = Thread(target=read_RTDE)
thread_read.start()


thread_main = Thread(target=main)
thread_main.start()

client.loop_forever()

#client.loop_stop()
